import os

#-------------------------------------------------------------------------
def DataPath():
    """ Return the default path for fit data h5 files"""
    return os.path.abspath('%s/data'%(os.path.dirname( \
        os.path.realpath(__file__))))



import numpy as np
from surfinBH import surfinBH

#=============================================================================
class Fit3dq8(surfinBH.SurFinBH):
    """ A class for the surfinBH3dq8 model presented in Varma et al., 2018,
    in prep. This model predicts the final mass mC, final spin chiC and final
    kick velocity velC, for the remnants of nonprecessing binary black hole
    systems. The fits are done using Gaussian Process Regression (GPR) and
    also provide an error estimate along with the fit value.

    This model has been trained in the parameter space:
        q <= 8, |chiAz| <= 0.8, |chiBz| <= 0.8

    However, it extrapolates reasonably to:
        q <= 10, |chiAz| <= 1, |chiBz| <= 1

    See __call__ method for evaluation.
    """

    #-------------------------------------------------------------------------
    def __init__(self, name):

        # soft_lims -> raise warning when outside lims
        # hard_lim -> raise error when outside lims
        # Same order as x in the call function. Each element is
        # a [minVal, maxVal] pair.
        soft_param_lims = [[0.99, 8.01], [-0.801, 0.801], [-0.801, 0.801]]
        hard_param_lims = [[0.99, 10.01], [-1, 1], [-1, 1]]
        super(Fit3dq8, self).__init__(name, soft_param_lims, hard_param_lims)


    #-------------------------------------------------------------------------
    def _load_fits(self, h5file):
        """ Loads fits from h5file and returns a dictionary of fits. """
        fits = {}
        for key in ['mC', 'chiCz', 'velCx', 'velCy']:
            fits[key] = self._load_scalar_fit(fit_key=key, h5file=h5file)
        return fits

    #-------------------------------------------------------------------------
    def _get_fit_params(self, x, fit_key):
        """ Transforms the input parameter to fit parameters for the 3dq8 model.
    That is, maps from [q, chiAz, chiBz] to [np.log(q), chiHat, chi_a]
    chiHat is defined in Eq.(3) of 1508.07253.
    chi_a = (chiAz - chiBz)/2.
        """
        q, chiAz, chiBz = x
        eta = q/(1.+q)**2
        chi_wtAvg = (q*chiAz+chiBz)/(1.+q)
        chiHat = (chi_wtAvg - 38.*eta/113.*(chiAz + chiBz))/(1. - 76.*eta/113.)
        chi_a = (chiAz - chiBz)/2.
        fit_params = [np.log(q), chiHat, chi_a]
        return fit_params


    #-------------------------------------------------------------------------
    def __call__(self, fit_key, x, **kwargs):
        """ Evaluates the surfinBH3dq8 model.

    Arguments:
        fit_key:
            'mC', 'chiC' or 'velC', for remnant mass, spin and kick velocity,
            respectively.

        x:
            Array of [q, chiAz, chiBz], where q >= 1 is the mass ratio and
            chiAz (chiBz) is the dimensionless spin of the larger (smaller) BH
            along the z-direction.

    Returns:
        If fit_key='mC':
            returns mC, mC_err_est
            The value and 1-sigma error estimate in remnant mass.
        If fit_key='chiC':
            returns chiCz, chiCz_err_est:
            The value and 1-sigma error estimate in remnant spin along the
            z-direction.
        If fit_key='velC':
            returns [velCx, velCy], [velCx_err_est, velCy_err_est]:
            The value and 1-sigma error estimate in the in-plane remnant kick
            vector.

    The spins and kick are defined in the coorbital frame at t=-100 M from the
    peak of the waveform. This frame is defined as:
    The z-axis is along the orbital angular momentum direction of the binary.
    The x-axis is along the line of separation from the smaller BH to
        the larger BH at this time.
    The y-axis completes the triad.
    We obtain this frame from the waveform as defined in arxiv:1705.07089.
        """
        # Warn/Exit if extrapolating
        self._check_param_limits(x)

        if fit_key == 'mC':
            mC, mC_err = self._evaluate_fits(x, 'mC')
            return mC, mC_err
        elif fit_key == 'chiC':
            chiCz, chiCz_err = self._evaluate_fits(x, 'chiCz')
            return chiCz, chiCz_err
        elif fit_key == 'velC':
            velCx, velCx_err = self._evaluate_fits(x, 'velCx')
            velCy, velCy_err = self._evaluate_fits(x, 'velCy')
            return np.array([velCx, velCy]), np.array([velCx_err, velCy_err])
        else:
            raise ValueError('Invalid fit_key')



"""surfinBH
========

Surrogate Final BH properties.

Example usage
-------------

To evaluate fits for remnant mass, spin and kick velocity of nonprecessing
binary black hole systems as presented in Varma:2018_in_prep.

import surfinBH

# First load the model
fit_3dq8 = surfinBH.LoadFits('3dq8')

# Mass ratio and component spins along orbital angular momentum direction
q = 6.7
chi1z = 0.74
chi2z = -0.6
x = [q, chi1z, chi2z]

## Evaluate the fits and GPR error estimate.

# Final mass
mC, mC_err_est = fit_3dq8('mC', x)

# Final spin along orbital angular momentum direction
chiCz, chiCz_err_est = fit_3dq8('chiC', x)

# Final kick velocity (velCz is zero for nonprecessing systems)
velC, velC_err_est = fit_3dq8('velC', x)
velCx, velCy = velC
velCx_err_est, velCy_err_est = velC_err_est

Additional examples can be found in the accompanying ipython notebooks.
To get a list of all available fits do:
surfinBH.FIT_CLASSES.keys()
"""

__copyright__ = "Copyright (C) 2018 Vijay Varma"
__email__ = "vvarma@caltech.edu"
__status__ = "testing"
__author__ = "Vijay Varma"
__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from _loadFits import LoadFits
from _loadFits import fits_collection
from _loadFits import DownloadData



import numpy as np
from surfinBH import surfinBH
from surfinBH._lal_spin_evolution import evolve_pn_spins
import surfinBH._utils as utils
import warnings

#=============================================================================
class Fit7dq2(surfinBH.SurFinBH):
    """ A class for the surfinBH7dq2 model presented in Varma et al., 2018,
    in prep. This model predicts the final mass mC, final spin vector chiC and
    final kick velocity vector velC, for the remnants of precessing binary
    black hole systems.  The fits are done using Gaussian Process Regression
    (GPR) and also provide an error estimate along with the fit value.

    This model has been trained in the parameter space:
        q <= 2, |chiA| <= 0.8, |chiB| <= 0.8

    However, it extrapolates reasonably to:
        q <= 3, |chiA| <= 1, |chiB| <= 1

    See __call__ method for evaluation.
    """

    #-------------------------------------------------------------------------
    def __init__(self, name, load_nrsur=False):

        #NOTE: These are not the actual limits.
        # We override _check_param_limits() to set limits, the
        # only purpose these serve is to generate regression data in
        # surfinBH/test/generate_regression_data.py
        soft_param_lims = None
        hard_param_lims = [[0.99, 3.01],
                [-0.6, 0.6],
                [-0.6, 0.6],
                [-0.6, 0.6],
                [-0.6, 0.6],
                [-0.6, 0.6],
                [-0.6, 0.6]]

        super(Fit7dq2, self).__init__(name, soft_param_lims, hard_param_lims)
        self.nrsur = None

    def _load_NRSur7dq2(self):
        import NRSur7dq2
        self.nrsur = NRSur7dq2.NRSurrogate7dq2()
        print('Loaded NRSur7dq2 waveform model')

    #-------------------------------------------------------------------------
    def _load_fits(self, h5file):
        """ Loads fits from h5file and returns a dictionary of fits. """
        fits = {}
        for key in ['mC']:
            fits[key] = self._load_scalar_fit(fit_key=key, h5file=h5file)
        for key in ['chiC', 'velC']:
            fits[key] = self._load_vector_fit(key, h5file)
        return fits

    #-------------------------------------------------------------------------
    def _get_fit_params(self, x, fit_key):
        """ Transforms the input parameter to fit parameters for the 7dq2 model.
    That is, maps from
    x = [q, chiAx, chiAy, chiAz, chiBx, chiBy, chiBz]
    fit_params = [np.log(q), chiAx, chiAy, chiHat, chiBx, chiBy, chi_a]

    chiHat is defined in Eq.(3) of 1508.07253, but with chiAz and chiBz instead
    of chiA and chiB.
    chi_a = (chiAz - chiBz)/2.
        """
        q, chiAz, chiBz = x[0], x[3], x[6]
        eta = q/(1.+q)**2
        chi_wtAvg = (q*chiAz+chiBz)/(1.+q)
        chiHat = (chi_wtAvg - 38.*eta/113.*(chiAz + chiBz))/(1. - 76.*eta/113.)
        chi_a = (chiAz - chiBz)/2.

        fit_params = x
        fit_params[0] = np.log(q)
        fit_params[3] = chiHat
        fit_params[6] = chi_a

        return fit_params

    #-------------------------------------------------------------------------
    def _check_param_limits(self, x):
        """ Checks that x is within allowed range of paramters.
        Raises a warning if outside training limits and
        raises an error if outside allowed limits.
        Training limits: q <= 2.01, chiAmag <= 0.81, chiBmag <= 0.81.
        Allowed limits: q <= 3.01, chiAmag <= 1, chiBmag <= 1.
        """
        q = x[0]
        chiAmag = np.sqrt(np.sum(x[1:4]**2))
        chiBmag = np.sqrt(np.sum(x[4:7]**2))

        if q < 1:
            raise ValueError('Mass ratio should be >= 1.')
        elif q > 3.01:
            raise Exception('Mass ratio outside allowed range.')
        elif q > 2.01:
            warnings.warn('Mass ratio outside training range.')

        if chiAmag > 1.:
            raise Exception('Spin magnitude of BhA outside allowed range.')
        elif chiAmag > 0.81:
            warnings.warn('Spin magnitude of BhA outside training range.')

        if chiBmag > 1.:
            raise Exception('Spin magnitude of BhB outside allowed range.')
        elif chiBmag > 0.81:
            warnings.warn('Spin magnitude of BhB outside training range.')

    def _evolve_spins(self, q, chiA0, chiB0, omega0, PN_approximant,
            PN_dt, PN_spin0, PN_phase0):
        """ Evolves spins of the component BHs from an initial orbital
        frequency = omega0 until t=-100 M from the peak of the waveform.
        If omega0 < 0.018, use PN to evolve the spins until
        orbital frequency = omega0. Then evolves further with the NRSur7dq2
        waveform model until t=-100M from the peak.

        Assumes chiA0 and chiB0 are defined in the inertial frame defined
        at orbital frequency = omega0 as:
            The z-axis is along the Newtonian orbital angular momentum when the
                PN orbital frequency = omega0.
            The x-axis is along the line of separation from the smaller BH to
                the larger BH at this frequency.
            The y-axis completes the triad.

        Returns spins in the coorbital frame at t=-100M, as well as the
        coprecessing frame quaternion and orbital phase in the coprecessing
        frame at this time.
        """

        # obrbital frequency beyond which we use the NRSur7dq2 model.
        omega0_nrsur = 0.018

        # If omega0 is below the NRSur7dq2 start frequency, we use PN
        # to evolve the spins until orbital frequency = omega0_nrsur.
        if omega0 < omega0_nrsur:
            # Note that we update omega0_nrsur here with the PN
            # frequency that was closest to the input omega0_nrsur.
            chiA0_nrsur_copr, chiB0_nrsur_copr, quat0_nrsur_copr, phi0_nrsur, \
                omega0_nrsur \
                = evolve_pn_spins(q, chiA0, chiB0, omega0,
                    omega0_nrsur, approximant=PN_approximant,
                    dt=PN_dt, spinO=PN_spin0,
                    phaseO=PN_phase0)

        # Load NRSur7dq2 if needed
        if self.nrsur is None:
            self._load_NRSur7dq2()

        # evaluate NRSur7dq2 dynamics
        quat, orbphase, chiA_copr, chiB_copr = self.nrsur.get_dynamics(q,
            chiA0_nrsur_copr, chiB0_nrsur_copr, init_quat=quat0_nrsur_copr,
            init_phase=phi0_nrsur, omega_ref=omega0_nrsur)

        # get data at time node where remnant fits are done
        fitnode_time = -100
        nodeIdx = np.argmin(np.abs(self.nrsur.tds - fitnode_time))
        quat_fitnode = quat.T[nodeIdx]
        orbphase_fitnode = orbphase[nodeIdx]

        # get coorbital frame spins at the time node
        chiA_coorb_fitnode = utils.rotate_in_plane(chiA_copr[nodeIdx],
                orbphase_fitnode)
        chiB_coorb_fitnode = utils.rotate_in_plane(chiB_copr[nodeIdx],
                orbphase_fitnode)

        return chiA_coorb_fitnode, chiB_coorb_fitnode, quat_fitnode, \
            orbphase_fitnode


    #-------------------------------------------------------------------------
    def __call__(self, fit_key, x, **kwargs):
        """Evaluates the surfinBH7dq2 model.

    Arguments:
        fit_key:
            'mC', 'chiC' or 'velC', for remnant mass, spin vector and kick
            vector respectively.

        x:
            Array of [q, chiAx, chiAy, chiAz, chiBx, chiBy, chiBz], where q >=
            1 is the mass ratio and chiA (chiB) is the dimensionless spin of
            the larger (smaller) BH.

            By default, the spins are assumed to be the component spins at
            t=-100 M from the peak of the waveform, and in the coorbital frame,
            defined as:
            The z-axis is along the orbital angular momentum at t=-100M.
            The x-axis is along the line of separation from the smaller BH to
                the larger BH at this time.
            The y-axis completes the triad.
            We obtain this frame from the waveform as defined in
            arxiv:1705.07089.

            If 'omega0' is given, instead the spins are assumed to be the
            component spins when the PN orbital frequency = omega0. The
            spins are assumed to be in the inertial frame, defined as:
            The z-axis is along the Newtonian orbital angular momentum when the
                PN orbital frequency = omega0.
            The x-axis is along the line of separation from the smaller BH to
                the larger BH at this frequency.
            The y-axis completes the triad.
            We obtain this frame from PN.
            Given the spins at omega0, we evolve the spins using PN until
            the orbital frequency = 0.018 M, then we further evolve the spins
            using the NRSur7dq2 model (arxiv:1705.07089) until t=-100M from the
            peak of the waveform.  Then we evaluate the fits using the spins at
            t=-100 M.  Finally, we transform the remnant spin and kick vectors
            back to the inertial frame defined above.

    Optional PN arguments:

        omega0:
            Initial dimensionless orbital frequency in units of 1/M, where M is
            the total mass. If this is given, the spins in x are assumed to be
            the component spins at this orbital frequency, and in the inertial
            frame as defined above. The returned remnant spin and kick vectors
            are also in the same frame.
            Default: None.

        PN_approximant:
            Approximant used to do the PN spin evolution. Choose from
            'SpinTaylorT1', 'SpinTaylorT4' or 'SpinTaylorT5'.
            Default: 'SpinTaylorT4'.

        PN_dt:
            Dimensionless time step size in units of M (total mass), used for
            the PN evolution. You may need to increase this if omega0 is very
            low.
            Default: 0.1

        PN_spin_order:
            Twice the PN order of spin effects. E.g., use 7 for 3.5PN.
            Default: 7

        PN_phase_order
            Twice the PN order in phase. E.g., use 7 for 3.5PN.
            Default: 7

    Returns:
        If fit_key='mC':
            returns mC, mC_err_est
            The value and 1-sigma error estimate in remnant mass.
        If fit_key='chiC':
            returns chiC, chiC_err_est
            The value and 1-sigma error estimate in remnant spin vector.
        If fit_key='velC':
            returns velC, velC_err_est
            The value and 1-sigma error estimate in remnant kick vector.

        By default, these vectors are defined in the coorbital frame at
            t=-100 M, as described above.
        If 'omega0' is not None, these are defined in the inertial frame at
            orbital frequency = omega0, as described above.
        """

        x = np.array(x)

        # Warn/Exit if extrapolating
        self._check_param_limits(x)

        omega0 = kwargs.pop('omega0', None)
        PN_approximant = kwargs.pop('PN_approximant', 'SpinTaylorT4')
        PN_dt = kwargs.pop('PN_dt', 0.1)
        PN_spin_order = kwargs.pop('PN_spin_order', 7)
        PN_phase_order = kwargs.pop('PN_phase_order', 7)

        # If omega0 is given, evolve the spins from omega0
        # to t = -100 M from the peak. Replace x with these spins.
        if omega0 is not None:
            q = x[0]
            chiA0 = x[1:4]
            chiB0 = x[4:7]
            chiA_coorb_fitnode, chiB_coorb_fitnode, quat_fitnode, \
                orbphase_fitnode \
                = self._evolve_spins(q, chiA0, chiB0, omega0,
                    PN_approximant, PN_dt, PN_spin_order,
                    PN_phase_order)

            # Update x to contain coorbital frame spins at t=-100M
            x = np.concatenate(([q], chiA_coorb_fitnode, chiB_coorb_fitnode))

        if fit_key == 'mC':
            fit_val, fit_err = self._evaluate_fits(x, fit_key)
        elif fit_key == 'chiC' or fit_key == 'velC':
            res = self._evaluate_fits(x, fit_key)
            fit_val = res.T[0]
            fit_err = res.T[1]
            if omega0 is not None:
                # If spins were given in inertial frame at omega0,
                # transform vectors and errors back to the same frame.
                fit_val = utils.transform_vector_coorb_to_inertial(fit_val,
                    orbphase_fitnode, quat_fitnode)
                fit_err = utils.transform_error_coorb_to_inertial(fit_val,
                    fit_err, orbphase_fitnode, quat_fitnode)
        else:
            raise ValueError('Invalid fit_key')

        if len(kwargs.keys()) != 0:
            unused = ""
            for k in kwargs.keys():
                unused += "'%s', "%k
            if unused[-2:] == ", ":     # get rid of trailing comma
                unused = unused[:-2]
            raise Exception('Unused keys in kwargs: %s'%unused)

        return fit_val, fit_err



__copyright__ = "Copyright (C) 2018 Vijay Varma"
__email__ = "vvarma@caltech.edu"
__status__ = "testing"
__author__ = "Vijay Varma"
__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import numpy as np
import os, sys
import h5py
import warnings

import _eval_pysur
from _dataPath import DataPath


#=============================================================================
class SurFinBH(object):
    """
Class to load and evaluate surrogate fits for final BH properties.
Each derived class should do the following:

    1. define _load_fits(self, h5file)
    2. define _get_fit_params(self, x, fit_key)
    3. define __call__(self, fit_key, x, **kwargs)
    4. define soft_param_lims and hard_param_lims, the limits for parameters
       beyond which warnings/errors are raised.

See _fit_evaluators.fit_7dq2.py for an example.
    """

    #-------------------------------------------------------------------------
    def __init__(self, name, soft_param_lims, hard_param_lims):
        """
        name:           Name of the fit excluding the surfinBH prefix. Ex: 7dq2.
        soft_param_lims: Parameter bounds beyond which a warning is raised.
        hard_param_lims: Parameter bounds beyond which an error is raised.
                         Same order and len as x in the __call__ function.
                         Each element is a [minVal, maxVal] pair.
                         Setting soft_param_lims/hard_param_lims to None will
                         skip that particular check.
        """
        self.name = name
        h5file = h5py.File('%s/fit_%s.h5'%(DataPath(), name), 'r')
        self.fits = self._load_fits(h5file)
        h5file.close()

        self.soft_param_lims = soft_param_lims
        self.hard_param_lims = hard_param_lims

    #-------------------------------------------------------------------------
    def _read_dict(self, f):
        """ Converts h5 groups to dictionaries
        """
        d = {}
        for k, item in f.iteritems():
            if type(item) == h5py._hl.dataset.Dataset:
                v = item.value
                if type(v) == np.string_:
                    v = str(v)
                if type(v) == str and v == "NONE":
                    d[k] = None
                elif type(v) == str and v == "EMPTYARR":
                    d[k] = np.array([])
                else:
                    d[k] = v
            elif k[:5] == "DICT_":
                d[k[5:]] = self._read_dict(item)
            elif k[:5] == "LIST_":
                tmpD = self._read_dict(item)
                d[k[5:]] = [tmpD[str(i)] for i in range(len(tmpD))]
        return d

    #-------------------------------------------------------------------------
    def _load_scalar_fit(self, fit_key=None, h5file=None, fit_data=None):
        """ Loads a single fit
        """
        if (fit_key is None) ^ (h5file is None):
            raise ValueError("Either specify both fit_key and h5file, or"
                " neither")

        if not ((fit_key is None) ^ (fit_data is None)):
            raise ValueError("Specify exactly one of fit_key and fit_data.")

        if fit_data is None:
            fit_data = self._read_dict(h5file[fit_key])

        if 'fitType' in fit_data.keys() and fit_data['fitType'] == 'GPR':
            fit = _eval_pysur.evaluate_fit.getGPRFitAndErrorEvaluator(fit_data)
        else:
            fit = _eval_pysur.evaluate_fit.getFitEvaluator(fit_data)

        return fit

    #-------------------------------------------------------------------------
    def _load_vector_fit(self, fit_key, h5file):
        """ Loads a vector of fits
        """
        vector_fit = []
        for i in range(len(h5file[fit_key].keys())):
            fit_data = self._read_dict(h5file[fit_key]['comp_%d'%i])
            vector_fit.append(self._load_scalar_fit(fit_data=fit_data))
        return vector_fit

    #-------------------------------------------------------------------------
    def _evaluate_fits(self, x, fit_key):
        """ Evaluates a particular fit by passing fit_key to self.fits.
            Assumes self._get_fit_params() has been overriden.
        """
        fit = self.fits[fit_key]
        fit_params = self._get_fit_params(x, fit_key)
        if type(fit) == list:
            res = []
            for i in range(len(fit)):
                res.append(fit[i](fit_params))
            return np.array(res)
        else:
            return fit(fit_params)

    #-------------------------------------------------------------------------
    def _check_param_limits(self, x):
        """ Checks that x is within allowed range of paramters.
            Raises a warning if outside self.soft_param_lims and
            raises an error if outside self.hard_param_lims.
            If these are None, skips the checks.
        """
        if self.soft_param_lims is not None:
            if len(x) != len(self.soft_param_lims):
                raise Exception("Expected x to be of len=%d"\
                    %len(self.soft_param_lims))

        if self.hard_param_lims is not None:
            for i in range(len(x)):
                if x[i] < self.hard_param_lims[i][0] \
                        or x[i] > self.hard_param_lims[i][1]:

                    raise Exception('Parameter x[%d] outside allowed range.'%i)

        if self.soft_param_lims is not None:
            for i in range(len(x)):
                if x[i] < self.soft_param_lims[i][0] \
                        or x[i] > self.soft_param_lims[i][1]:

                    warnings.warn('Parameter x[%d] outside training range.'%i)

    #-------------------------------------------------------------------------
    #----------------------  Override these  ---------------------------------
    #-------------------------------------------------------------------------

    #-------------------------------------------------------------------------
    def _load_fits(self, h5file):
        """ Loads fits from h5file and returns a dictionary of fits. """
        raise NotImplementedError("Please override me.")
        return fits

    #-------------------------------------------------------------------------
    def _get_fit_params(self, x, fit_key):
        """ Maps from input params x to the fit_params used to evaluate the
            fit.
        """
        raise NotImplementedError("Please override me.")
        return fit_params

    #-------------------------------------------------------------------------
    def __call__(self, fit_key, x, **kwargs):
        """ Evaluates a particular fit. This varies for each surrogate.
            Allowed values for fit_key are 'mC', 'chiC' and 'velC'.
            If fit_key is a scalar, returns the fit value and error
            estimate (if available).
            If fit_key is a scalar, returns two arrays. The first array
            is a vector of fit values and the second is the corresponding
            vector of errors estimates (if available).

            Each derived class should have its own __call__ function but
            call self._check_param_limits() first to do some sanity checks.
            See _fit_evaluators.fit_7dq2.py for an example.
        """
        raise NotImplementedError("Please override me.")



import numpy as np
import h5py
import os
import warnings

import surfinBH

# Ignore these specific warning.
warnings.filterwarnings("ignore", message="Parameter x[[0-9]+] outside"
        " training range.")
warnings.filterwarnings("ignore", message="Mass ratio outside training range.")
warnings.filterwarnings("ignore", message="Spin magnitude of BhA outside"
        " training range.")
warnings.filterwarnings("ignore", message="Spin magnitude of BhB outside"
        " training range.")


def test_fit_regression():
    """ Compares all existing fits against saved regression data.
        Regression data should already have been generated by
        generate_regression_data.py
    """
    # List of all available fits
    fit_names = surfinBH.fits_collection.keys()
    for name in fit_names:
        short_name = name.split('surfinBH')[-1]

        # Download fit data if needed
        if not os.path.isfile('%s/fit_%s.h5'%(surfinBH._dataPath.DataPath(),
                short_name)):

            surfinBH.DownloadData(name)


        # Load fit
        fit = surfinBH.LoadFits(name)

        # Load regression data
        regression_h5file = h5py.File('test/regression_data/fit_%s.h5'%(
                short_name))

        # Compare fit and regression data for each regression test
        test_keys = regression_h5file.keys()
        for test in test_keys:
            print('... running %s'%test)
            test_h5grp = regression_h5file[test]
            x = test_h5grp['x'].value
            fit_keys = test_h5grp['y'].keys()
            for key in fit_keys:
                print('...... testing %s'%key)
                y_reg = test_h5grp['y/%s'%key].value
                y_fit = fit(key, x)
                # lower rtol seems to make the test fail for the GPR error
                # estimates when testing on a different machine. This seems to
                # happen even with the same version of sklearn.
                np.testing.assert_allclose(y_fit, y_reg, rtol=1e-6)

        regression_h5file.close()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="surfinBH",
    version="0.0.8.dev0",
    author="Vijay Varma",
    author_email="vvarma@caltech.edu",
    description="Surrogate Final BH properties.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vijayvarma392/surfinBH",
    packages=setuptools.find_packages(),
    classifiers=[
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Physics",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Operating System :: OS Independent",
    ],
)



import os
from time import gmtime, strftime
from glob import glob

import _fit_evaluators
from _dataPath import DataPath

#=============================================================================
class FitAttributes(object):
    """ Saves attributes of a particular fit.
    """
    #-------------------------------------------------------------------------
    def __init__(self, **kwargs):
        self.fit_class =  kwargs['fit_class']
        self.desc = kwargs['desc']
        self.data_url =  kwargs['data_url']
        self.refs =  kwargs['refs']

#-------------------------------------------------------------------------
def LoadFits(name):
    if name not in fits_collection.keys():
        raise Exception('Invalid fit name : %s'%name)
    else:
        fit = fits_collection[name].fit_class(name.split('surfinBH')[-1])
        print('Loaded %s fit.'%name)
        return fit

#-------------------------------------------------------------------------
def DownloadData(name, data_dir=DataPath() ):
    """ Downloads fit data to DataPath() diretory.
    """
    if name not in fits_collection.keys():
        raise Exception('Invalid fit name : %s'%name)

    data_url = fits_collection[name].data_url
    fname = os.path.basename(data_url)

    # If file already exists, move it to backup dir with time stamp
    if os.path.isfile('%s/%s'%(data_dir, fname)):
        timestamp=strftime("%Y%b%d_%Hh:%Mm:%Ss", gmtime())
        backup_fname = '%s_%s'%(timestamp, fname)
        backup_dir = '%s/backup'%(data_dir)
        os.system('mkdir -p %s'%backup_dir)
        print('\n%s file exits, moving to %s/%s.'%(fname, backup_dir, \
            backup_fname))
        os.system('mv %s/%s %s/%s'%(data_dir, fname, backup_dir, backup_fname))
        number_of_backup_files = glob('%s/*_%s'%(backup_dir, fname))
        if len(number_of_backup_files) > 5:
            print('There are a lot of backup files in %s, consider removing'
                ' some.'%backup_dir)

    os.system('mkdir -p {0}; cd {0}; curl -# -L -O {1}'.format(data_dir,
            data_url))


##############################################################################

#### Add fits here, each name should start with surfinBH.
fits_collection = {}

fits_collection['surfinBH3dq8'] = FitAttributes( \
    fit_class = _fit_evaluators.Fit3dq8,
    desc = 'Fits for remnant mass, spin and kick veclocity for nonprecessing'
        ' BBH systems.',
    data_url = 'https://www.dropbox.com/s/06mrxalxqjhzy9d/fit_3dq8.h5',
    refs = 'arxiv.2018.xxxx',
    )

fits_collection['surfinBH7dq2'] = FitAttributes( \
    fit_class = _fit_evaluators.Fit7dq2,
    desc = 'Fits for remnant mass, spin and kick veclocity for genrically'
        ' precessing BBH systems.',
    data_url = 'https://www.dropbox.com/s/8b4o7n5aswrnami/fit_7dq2.h5',
    refs = 'arxiv.2018.xxxx',
    )
